@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -16,7 +16,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #ifndef SL_CONFIG_H
@@ -45,7 +45,6 @@ USA.
 # define __BORLAND_V5__
 #endif
 
-
 #if defined(__WATCOMC__) && defined(__DOS__)
 # define  DOS386
 #endif
@@ -158,7 +157,6 @@ USA.
 # define HAVE_STRTOD	1
 #endif
 
-
 /* Does VMS have this??? */
 #if !defined(VMS) && !defined(_MSC_VER) && (!defined(__WATCOMC__) || defined(__QNX__))
 # define HAVE_DIRENT_H	1
@@ -196,6 +194,10 @@ USA.
 #undef HAVE_GETTIMEOFDAY
 
 #undef HAVE_GMTIME
+#undef HAVE_GMTIME_R
+#undef HAVE_LOCALTIME_R
+#undef HAVE_CTIME_R
+
 #ifdef __unix__
 # define HAVE_GMTIME	1
 #endif
@@ -206,6 +208,12 @@ USA.
 #undef HAVE_READLINK
 #undef HAVE_UNAME
 
+#if defined(__MINGW32__)
+# define HAVE_UTIME
+# define HAVE_UTIME_H
+#endif
+#undef HAVE_UTIMES
+
 #undef HAVE_POPEN
 #if defined(__MINGW32__)
 # define HAVE_POPEN 1
@@ -263,13 +271,13 @@ USA.
 /* #define gid_t int */
 /* #define off_t int */
 /* #define size_t unsigned int */
+#define SIZEOF_OFF_T	4
+#define SIZEOF_SIZE_T	4
 
 #undef HAVE_LONG_LONG
 
 #define HAVE_SETVBUF
 
-#define SIZEOF_OFF_T	4
-
 #if defined (__EMX__) || defined(__BORLANDC__) || defined(__IBMC__) || defined(_MSC_VER) || defined(__WATCOMC__)
 # define HAVE_IO_H
 #endif
@@ -324,6 +332,8 @@ USA.
 
 #undef HAVE_FORK
 #undef HAVE_WAITPID
+#undef HAVE_TTYNAME
+#undef HAVE_TTYNAME_R
 
 /* Files used by the slang-readline interface.  The interface first tries
  * to read SLRLINE_USER_INIT_FILE from $HOME, and if that does not exist, it
