@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -16,7 +16,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 /* Support for UTF-8 combining characters added by Adrian Colley */
@@ -162,14 +162,14 @@ static int getkey_function (void)
 	if (Keyboard_Buffer_Stop == Keyboard_Buffer + sizeof (Keyboard_Buffer))
 	  Keyboard_Buffer_Stop = Keyboard_Buffer;
      }
-   
+
    return ch;
 }
 
 static int get_buffered_key (void)
 {
    int ch;
-   
+
    if (Keyboard_Buffer_Stop == Keyboard_Buffer_Start)
      return SLANG_GETKEY_ERROR;
 
@@ -183,7 +183,7 @@ static int get_buffered_key (void)
 static int get_keypad_key (void)
 {
    int ch;
-   
+
    if (Keyboard_Buffer_Stop != Keyboard_Buffer_Start)
      return get_buffered_key ();
 
@@ -505,7 +505,7 @@ SLcurses_Window_Type *SLcurses_newwin (unsigned int nrows, unsigned int ncols,
    if (ncols == 0)
      ncols = (unsigned int) SLtt_Screen_Cols - c;
 
-   lines = (SLcurses_Cell_Type **) SLmalloc (nrows * sizeof (SLcurses_Cell_Type *));
+   lines = (SLcurses_Cell_Type **) _SLcalloc (nrows, sizeof (SLcurses_Cell_Type *));
    if (lines == NULL)
      {
 	SLcurses_delwin (win);
@@ -528,7 +528,7 @@ SLcurses_Window_Type *SLcurses_newwin (unsigned int nrows, unsigned int ncols,
      {
 	SLcurses_Cell_Type *b;
 
-	b = (SLcurses_Cell_Type *) SLmalloc (ncols * sizeof (SLcurses_Cell_Type));
+	b = (SLcurses_Cell_Type *) _SLcalloc (ncols, sizeof (SLcurses_Cell_Type));
 	if (b == NULL)
 	  {
 	     SLcurses_delwin (win);
@@ -982,7 +982,7 @@ int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
    while (u < umax)
      {
 	SLwchar_Type ch;
-	unsigned int nconsumed;
+	SLstrlen_Type nconsumed;
 	int width = 1;
 
 	if (SLsmg_is_utf8_mode () && SLutf8_decode (u, umax, &ch, &nconsumed))
@@ -1125,7 +1125,7 @@ SLcurses_Window_Type *SLcurses_subwin (SLcurses_Window_Type *orig,
    sw->_maxx = (begin_x + ncols) - 1;
    sw->_maxy = (begin_y + nlines) - 1;
 
-   sw->lines = (SLcurses_Cell_Type **) SLmalloc (nlines * sizeof (SLcurses_Cell_Type *));
+   sw->lines = (SLcurses_Cell_Type **) _SLcalloc (nlines, sizeof (SLcurses_Cell_Type *));
    if (sw->lines == NULL)
      {
 	SLcurses_delwin (sw);
