@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -16,7 +16,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #include "slinclud.h"
@@ -76,6 +76,16 @@ USA.
 #include "slang.h"
 #include "_slang.h"
 
+#define CROSS_COMPILE_FOR_VMS	0
+#ifdef VMS
+# define VMS_SYSTEM 1
+#else
+# if CROSS_COMPILE_FOR_VMS
+#  define VMS 1
+#  undef __unix__
+# endif
+#endif
+
 /* Colors:  These definitions are used for the display.  However, the
  * application only uses object handles which get mapped to this
  * internal representation.  The mapping is performed by the Color_Map
@@ -104,7 +114,10 @@ int SLtt_Blink_Mode = 0;
 int SLtt_Use_Blink_For_ACS = 0;
 int SLtt_Newline_Ok = 0;
 int SLtt_Has_Alt_Charset = 0;
-int SLtt_Force_Keypad_Init = 0;
+int SLtt_Force_Keypad_Init = -1;
+
+static int Use_Relative_Cursor_Addressing = 0;
+static int Max_Relative_Cursor_r = 0;
 
 /* static int UTF8_Mode = -1; */
 
@@ -117,6 +130,7 @@ static int Can_Background_Color_Erase = 1;
 int SLtt_Has_Status_Line = -1;	       /* hs */
 int SLang_TT_Write_FD = -1;
 
+static int TT_Is_Initialized = 0;
 static int Automatic_Margins;
 /* static int No_Move_In_Standout; */
 static int Worthless_Highlight;
@@ -164,7 +178,6 @@ static SLCONST int RGB_to_BGR[] =
      0, 4, 2, 6, 1, 5, 3, 7
 };
 
-
 static SLCONST char *Color_Fg_Str = "\033[3%dm";
 static SLCONST char *Color_Bg_Str = "\033[4%dm";
 static SLCONST char *Default_Color_Fg_Str = "\033[39m";
@@ -192,8 +205,16 @@ static SLCONST char *Del_Char_Str; /* = "\033[P"; */   /* dc */
 static SLCONST char *Del_N_Lines_Str; /* = "\033[%dM"; */  /* DL */
 static SLCONST char *Add_N_Lines_Str; /* = "\033[%dL"; */  /* AL */
 static SLCONST char *Rev_Scroll_Str;
-static SLCONST char *Curs_Up_Str;
-static SLCONST char *Curs_F_Str;    /* RI termcap string */
+static SLCONST char *Curs_Up_Str;      /* "up" */
+static SLCONST char *Curs_UpN_Str;     /* "UP" */
+static SLCONST char *Curs_Dn_Str;      /* "do" */
+static SLCONST char *Curs_DnN_Str;     /* "DO" */
+static SLCONST char *Curs_Right_Str;    /* "nd" */
+static SLCONST char *Curs_RightN_Str;    /* "RI" */
+static SLCONST char *Curs_Left_Str;    /* "bc", "le" */
+static SLCONST char *Curs_LeftN_Str;    /* "LE" */
+static SLCONST char *Clear_EOS_Str;   /* cd */
+
 static SLCONST char *Cursor_Visible_Str;    /* ve termcap string */
 static SLCONST char *Cursor_Invisible_Str;    /* vi termcap string */
 #if 0
@@ -204,23 +225,21 @@ static SLCONST char *Start_Alt_Chars_Str;  /* as */
 static SLCONST char *End_Alt_Chars_Str;   /* ae */
 static SLCONST char *Enable_Alt_Char_Set;  /* eA */
 
-static SLCONST char *Term_Init_Str;
+static SLCONST char *Start_Abs_Cursor_Addressing_Mode;
 static SLCONST char *Keypad_Init_Str;
-static SLCONST char *Term_Reset_Str;
+static SLCONST char *End_Abs_Cursor_Addressing_Mode;
 static SLCONST char *Keypad_Reset_Str;
 
 /* status line functions */
 static SLCONST char *Disable_Status_line_Str;  /* ds */
 static SLCONST char *Return_From_Status_Line_Str;   /* fs */
 static SLCONST char *Goto_Status_Line_Str;     /* ts */
-static int Num_Status_Line_Columns;    /* ws */
+/* static int Num_Status_Line_Columns; */   /* ws */
 /* static int Status_Line_Esc_Ok;	 */       /* es */
 
-/* static int Len_Curs_F_Str = 5; */
-
 /* cm string has %i%d since termcap numbers columns from 0 */
 /* char *CURS_POS_STR = "\033[%d;%df";  ansi-- hor and vert pos */
-static SLCONST char *Curs_Pos_Str; /* = "\033[%i%d;%dH";*/   /* cm termcap string */
+static SLCONST char *Abs_Curs_Pos_Str; /* = "\033[%i%d;%dH";*/   /* cm termcap string */
 
 /* scrolling region */
 static int Scroll_r1 = 0, Scroll_r2 = 23;
@@ -232,6 +251,8 @@ static SLtt_Char_Type Current_Fgbg = 0xFFFFFFFFU;
 static int Cursor_Set;		       /* 1 if cursor position known, 0
 					* if not.  -1 if only row is known
 					*/
+/* This is used only in relative-cursor-addressing mode */
+static SLsmg_Char_Type Display_Start_Chars[SLTT_MAX_SCREEN_ROWS];
 
 #define MAX_OUTPUT_BUFFER_SIZE 4096
 
@@ -680,6 +701,9 @@ static void tt_printf(SLCONST char *fmt, int x, int y)
 
 void SLtt_set_scroll_region (int r1, int r2)
 {
+   if (Use_Relative_Cursor_Addressing)
+     return;
+
    Scroll_r1 = r1;
    Scroll_r2 = r2;
    tt_printf (Scroll_R_Str, Scroll_r1, Scroll_r2);
@@ -700,6 +724,70 @@ int SLtt_set_cursor_visibility (int show)
    return 0;
 }
 
+static void cursor_motion (SLCONST char *s1, SLCONST char *sN, int n)
+{
+   if ((n == 1) && (s1 != NULL))
+     {
+	tt_write_string (s1);
+	return;
+     }
+
+   if (n <= 0)
+     return;
+
+   if (sN != NULL)
+     tt_printf (sN, n, 0);
+   else while (n > 0)
+     {
+	tt_write_string (s1);
+	n--;
+     }
+}
+
+static void goto_relative_rc (int r, int c)
+{
+   if (r < 0)
+     return;
+
+   if (Cursor_Set != 1)
+     {
+	/* Do not know where the cursor is.  At least get the column correct */
+	tt_write ("\r", 1);
+	Cursor_c = 0;
+	Cursor_Set = 1;		       /* pretend we know the row */
+     }
+
+   if (Cursor_r > r)
+     cursor_motion (Curs_Up_Str, Curs_UpN_Str, Cursor_r-r);
+   else if (Cursor_r < r)
+     {
+	tt_write ("\r", 1); Cursor_c = 0;
+	if (r > Max_Relative_Cursor_r)
+	  {
+	     cursor_motion (Curs_Dn_Str, Curs_DnN_Str, Max_Relative_Cursor_r-Cursor_r);
+	     Cursor_r = Max_Relative_Cursor_r;
+	     while (Cursor_r < r)
+	       {
+		  tt_write ("\n", 1);
+		  Cursor_r++;
+	       }
+	  }
+	else
+	  cursor_motion (Curs_Dn_Str, Curs_DnN_Str, r-Cursor_r);
+     }
+   Cursor_r = r;
+   if (r > Max_Relative_Cursor_r)
+     Max_Relative_Cursor_r = r;
+
+   if (Cursor_c > c)
+     cursor_motion (Curs_Left_Str, Curs_LeftN_Str, Cursor_c-c);
+   else if (Cursor_c < c)
+     cursor_motion (Curs_Right_Str, Curs_RightN_Str, c-Cursor_c);
+
+   Cursor_c = c;
+   Cursor_Set = 1;
+}
+
 /* the goto_rc function moves to row relative to scrolling region */
 void SLtt_goto_rc(int r, int c)
 {
@@ -710,13 +798,21 @@ void SLtt_goto_rc(int r, int c)
    if ((c < 0) || (r < 0))
      {
 	Cursor_Set = 0;
+	Cursor_c = 0;
+	Cursor_r = 0;
+	tt_write("\r", 1);
+	return;
+     }
+   if (Use_Relative_Cursor_Addressing)
+     {
+	goto_relative_rc (r, c);
 	return;
      }
 
-   /* if (No_Move_In_Standout && Current_Fgbg) SLtt_normal_video (); */
    r += Scroll_r1;
 
-   if ((Cursor_Set > 0) || ((Cursor_Set < 0) && !Automatic_Margins))
+   if ((Cursor_Set > 0)
+       || ((Cursor_Set < 0) && !Automatic_Margins))
      {
 	n = r - Cursor_r;
 	if ((n == -1) && (Cursor_Set > 0) && (Cursor_c == c)
@@ -765,7 +861,10 @@ void SLtt_goto_rc(int r, int c)
 	  }
      }
    if (s != NULL) tt_write_string(s);
-   else tt_printf(Curs_Pos_Str, r, c);
+   else
+     {
+	tt_printf(Abs_Curs_Pos_Str, r, c);
+     }
    Cursor_c = c; Cursor_r = r;
    Cursor_Set = 1;
 }
@@ -852,9 +951,9 @@ void SLtt_delete_nlines (int nn)
      }
 }
 
-void SLtt_cls (void)
+static void cls_internal (SLCONST char *escseq, int rmin)
 {
-   /* If the terminal is a color terminal but the user wants black and 
+   /* If the terminal is a color terminal but the user wants black and
     * white, then make sure that the colors are reset.  This appears to be
     * necessary.
     */
@@ -868,7 +967,27 @@ void SLtt_cls (void)
 
    SLtt_normal_video();
    SLtt_reset_scroll_region ();
-   tt_write_string(Cls_Str);
+
+   tt_write_string(escseq);
+
+   if (Use_Relative_Cursor_Addressing)
+     {
+	int r, rmax = SLtt_Screen_Rows;
+	for (r = rmin; r < rmax; r++)
+	  Display_Start_Chars[r].nchars = 0;
+     }
+}
+
+void SLtt_cls (void)
+{
+   cls_internal (Cls_Str, 0);
+}
+
+static void _pSLtt_clear_eos (void)
+{
+   int rmin = Cursor_r;
+   if (Cursor_c > 0) rmin++;
+   cls_internal (Clear_EOS_Str, rmin);
 }
 
 void SLtt_reverse_index (int n)
@@ -917,9 +1036,13 @@ static void write_string_with_care (SLCONST char *);
 
 static void del_eol (void)
 {
-#if 0
-   int c;
-#endif
+   if ((Cursor_c == 0)
+       && (Use_Relative_Cursor_Addressing)
+       && (Cursor_r < SLTT_MAX_SCREEN_ROWS))
+     {
+	Display_Start_Chars[Cursor_r].nchars = 0;
+     }
+
    if ((Del_Eol_Str != NULL)
        && (Can_Background_Color_Erase || ((Current_Fgbg & ~0xFF) == 0)))
      {
@@ -927,23 +1050,6 @@ static void del_eol (void)
 	return;
      }
 
-#if 0
-   c = Cursor_c;
-   
-   /* Avoid writing to the lower right corner.  If the terminal does not
-    * have Del_Eol_Str, then it probably does not have what it takes to play
-    * games with insert-mode to "push" the desired character into that corner.
-    */
-   if (Cursor_r + 1 < SLtt_Screen_Rows)
-     c++;
-
-   while (c < SLtt_Screen_Cols)
-     {
-	tt_write (" ", 1);
-	c++;
-     }
-   Cursor_c = (SLtt_Screen_Cols-1);
-#else
    while (Cursor_c < SLtt_Screen_Cols)
      {
 	write_string_with_care (" ");
@@ -951,7 +1057,6 @@ static void del_eol (void)
      }
    Cursor_c = SLtt_Screen_Cols - 1;
    Cursor_Set = 0;
-#endif
 }
 
 void SLtt_del_eol (void)
@@ -1025,8 +1130,7 @@ static int initialize_brushes (void)
    Brushes_Initialized = 1;
    return 0;
 }
-   
-   
+
 static Brush_Info_Type *get_brush_info (SLsmg_Color_Type color)
 {
    if (Brushes_Initialized == 0)
@@ -1043,13 +1147,13 @@ static Brush_Info_Type *get_brush_info (SLsmg_Color_Type color)
 static SLtt_Char_Type get_brush_attr (SLsmg_Color_Type color)
 {
    Brush_Info_Type *b;
-   
+
    if (NULL == (b = get_brush_info (color)))
      return (SLtt_Char_Type)-1;
-   
+
    if (SLtt_Use_Ansi_Colors)
      return b->fgbg;
-   
+
    return b->mono;
 }
 
@@ -1065,7 +1169,7 @@ int SLtt_set_mono (int obj, SLFUTURE_CONST char *what, SLtt_Char_Type mask)
    (void) what;
    if (NULL == (b = get_brush_info (obj)))
      return -1;
-   
+
    b->mono = mask & ATTR_MASK;
    return 0;
 }
@@ -1162,7 +1266,7 @@ static int Color_0_Modified = 0;
 int SLtt_set_color_object (int obj, SLtt_Char_Type attr)
 {
    Brush_Info_Type *b;
-   
+
    if (NULL == (b = get_brush_info (obj)))
      return -1;
 
@@ -1171,7 +1275,7 @@ int SLtt_set_color_object (int obj, SLtt_Char_Type attr)
 
    if (_pSLtt_color_changed_hook != NULL)
      (*_pSLtt_color_changed_hook)();
-   
+
    return 0;
 }
 
@@ -1183,7 +1287,7 @@ SLtt_Char_Type SLtt_get_color_object (int obj)
 int SLtt_add_color_attribute (int obj, SLtt_Char_Type attr)
 {
    Brush_Info_Type *b;
-   
+
    if (NULL == (b = get_brush_info (obj)))
      return -1;
 
@@ -1192,7 +1296,7 @@ int SLtt_add_color_attribute (int obj, SLtt_Char_Type attr)
    if (obj == 0) Color_0_Modified = 1;
    if (_pSLtt_color_changed_hook != NULL)
      (*_pSLtt_color_changed_hook)();
-   
+
    return 0;
 }
 
@@ -1254,7 +1358,7 @@ static int parse_color_digit_name (SLCONST char *color, SLtt_Char_Type *f)
 	  break;
 	if ((ch > '9') || (ch < '0'))
 	  return -1;
-	
+
 	if (i > 0xFFFFFFFFU / 10)
 	  return -1;
 	j = (i *= 10);
@@ -1317,7 +1421,7 @@ static int tt_set_color (int obj, SLCONST char *what, SLCONST char *fg, SLCONST
    SLtt_Char_Type fgbg;
 
    (void) what;
-   
+
    if (-1 == make_color_fgbg (fg, bg, &fgbg))
      return -1;
 
@@ -1329,7 +1433,6 @@ int SLtt_set_color (int obj, SLFUTURE_CONST char *what, SLFUTURE_CONST char *fg,
    return tt_set_color (obj, what, fg, bg);
 }
 
-
 int SLtt_set_color_fgbg (int obj, SLtt_Char_Type f, SLtt_Char_Type b)
 {
    return SLtt_set_color_object (obj, fb_to_fgbg (f, b));
@@ -1393,7 +1496,7 @@ static void write_attributes (SLtt_Char_Type fgbg)
 	fg0 = (int) GET_FG(fgbg);
 	bg0 = (int) GET_BG(fgbg);
 
-	if (unknown_attributes 
+	if (unknown_attributes
 	    || (fg0 != (int)GET_FG(Current_Fgbg)))
 	  {
 	     if (fg0 == SLSMG_COLOR_DEFAULT)
@@ -1467,14 +1570,14 @@ static int bce_colors_eq (SLsmg_Color_Type ca, SLsmg_Color_Type cb, int just_bg)
 
    ba = get_brush_info (ca);
    bb = get_brush_info (cb);
-   
+
    if (SLtt_Use_Ansi_Colors == 0)
      return ba->mono == bb->mono;
 
    if (Bce_Color_Offset)
      {
 	/* If either are color 0, then we do not know what that means since the
-	 * terminal does not support BCE 
+	 * terminal does not support BCE
 	 */
 	if ((ca == 0) || (cb == 0))
 	  return 0;
@@ -1485,7 +1588,10 @@ static int bce_colors_eq (SLsmg_Color_Type ca, SLsmg_Color_Type cb, int just_bg)
    if (ba->fgbg == bb->fgbg)
      return 1;
    if (just_bg)
-     return GET_BG(ba->fgbg) == GET_BG(bb->fgbg);
+     {
+	return (ba->mono == bb->mono)
+	  && GET_BG(ba->fgbg) == GET_BG(bb->fgbg);
+     }
    return 0;
 }
 
@@ -1501,7 +1607,7 @@ static void write_string_with_care (SLCONST char *str)
    len = strlen (str);
    if (Automatic_Margins && (Cursor_r + 1 == SLtt_Screen_Rows))
      {
-	if (_pSLtt_UTF8_Mode == 0) 
+	if (_pSLtt_UTF8_Mode == 0)
 	 {
 	   if (len + (unsigned int) Cursor_c >= (unsigned int) SLtt_Screen_Cols)
 	     {
@@ -1510,13 +1616,13 @@ static void write_string_with_care (SLCONST char *str)
 	        */
 	        if (SLtt_Screen_Cols > Cursor_c)
 	          len = SLtt_Screen_Cols - Cursor_c - 1;
-	        else 
+	        else
 	          len = 0;
 	     }
 	 }
        else
 	 {
-	    unsigned int nchars = SLutf8_strlen((SLuchar_Type *)str, 1);
+	    SLstrlen_Type nchars = SLutf8_strlen((SLuchar_Type *)str, 1);
 	    if (nchars + (unsigned int) Cursor_c >= (unsigned int) SLtt_Screen_Cols)
 	     {
 	       if (SLtt_Screen_Cols > Cursor_c)
@@ -1527,7 +1633,7 @@ static void write_string_with_care (SLCONST char *str)
 		    len = p - str;
 		 }
 	       else
-		  len = 0; 
+		  len = 0;
 	     }
 	 }
      }
@@ -1545,6 +1651,16 @@ static void send_attr_str (SLsmg_Char_Type *s, SLsmg_Char_Type *smax)
    p = out;
    pmax = p + (sizeof (out)-1);
 
+   if ((Cursor_c == 0)
+       && (Use_Relative_Cursor_Addressing)
+       && (Cursor_r < SLTT_MAX_SCREEN_ROWS))
+     {
+	if (s < smax)
+	  Display_Start_Chars[Cursor_r] = *s;
+	else
+	  Display_Start_Chars[Cursor_r].nchars = 0;
+     }
+
    dcursor_c = 0;
    while (s < smax)
      {
@@ -1607,7 +1723,7 @@ static void send_attr_str (SLsmg_Char_Type *s, SLsmg_Char_Type *smax)
 		    }
 	       }
 	  }
-	
+
 	if ((wch < 0x80) && (nchars == 1))
 	  *p++ = (unsigned char) wch;
 	else if (_pSLtt_UTF8_Mode == 0)
@@ -1642,7 +1758,6 @@ static void forward_cursor (unsigned int n, int row)
 {
    char buf [1024];
 
-
    /* if (Current_Fgbg & ~0xFF) */
    /*   { */
    /* 	unsigned int num = 0; */
@@ -1654,8 +1769,7 @@ static void forward_cursor (unsigned int n, int row)
    /* 	Cursor_c += n; */
    /* 	return; */
    /*   } */
-   
-	
+
    if (n <= 4)
      {
 	SLtt_normal_video ();
@@ -1668,17 +1782,16 @@ static void forward_cursor (unsigned int n, int row)
 	write_string_with_care (buf);
 	Cursor_c += n;
      }
-   else if (Curs_F_Str != NULL)
+   else if (Curs_RightN_Str != NULL)
      {
 	Cursor_c += n;
-	n = tt_sprintf(buf, sizeof (buf), Curs_F_Str, (int) n, 0);
+	n = tt_sprintf(buf, sizeof (buf), Curs_RightN_Str, (int) n, 0);
 	tt_write(buf, n);
      }
    else SLtt_goto_rc (row, (int) (Cursor_c + n));
 }
 
-
-/* FIXME!!  If the terminal does not support color, then this route has 
+/* FIXME!!  If the terminal does not support color, then this route has
  * problems of color object 0 has been assigned some monochrome attribute
  * such as reverse video.  In such a case, space_char=' ' is not a simple
  * space character as is assumed below.
@@ -1734,14 +1847,14 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 
 	a = oldd+(len-1);
 	b = neww+(len-1);
-	
+
 	if (CHAR_EQS(a,b))
 	  insert_hack_char = NULL;
-	else 
+	else
 	  insert_hack_prev = &neww[len-2];
      }
 #endif
-     
+
    memset ((char *) &space_char_buf, 0, sizeof (SLsmg_Char_Type));
    space_char = &space_char_buf;
    space_char->nchars = 1;
@@ -1819,15 +1932,21 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
    if (((pmax-1)->wchars[0]==' ') && ((pmax-1)->nchars==1))
      {
 	/* If we get here, then we can erase to the end of the line to create
-	 * the final space.  However, this will only work _if_ erasing will 
+	 * the final space.  However, this will only work _if_ erasing will
 	 * get us the correct color.  If the terminal supports BCE, then this
 	 * is easy.  If it does not, then we can only perform this operation
-	 * if the color is known via something like COLORFGBG.  For now, 
+	 * if the color is known via something like COLORFGBG.  For now,
 	 * I just will not perform the optimization for such terminals.
 	 */
-	if ((Can_Background_Color_Erase)
+	if (Can_Background_Color_Erase
 	    && SLtt_Use_Ansi_Colors)
-	  space_char = pmax - 1;
+	  {
+	     SLtt_Char_Type fgbg;
+
+	     fgbg = get_brush_attr (COLOR_OF(pmax-1));
+	     if (0 == (fgbg & ATTR_MASK))
+	       space_char = pmax - 1;
+	  }
 
 	while (pmax > p)
 	  {
@@ -1887,12 +2006,12 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 		  while ((p1 < pmax) && (CHAR_EQS_SPACE (p1)))
 		    p1++;
 	       }
-	     
+
 	     /* Is this optimization worth it?  Assume Del_Bol_Str is ESC [ 1 K
-	      * It costs 4 chars + the space needed to properly position the 
+	      * It costs 4 chars + the space needed to properly position the
 	      * cursor, e.g., ESC [ 10;10H. So, it costs at least 13 characters.
 	      */
-	     if ((p1 > neww + 13) 
+	     if ((p1 > neww + 13)
 		 && (p1 >= p)
 		 /* Avoid erasing from the end of the line */
 		 && ((p1 != pmax) || (pmax < neww + len)))
@@ -1914,8 +2033,7 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 #ifdef HP_GLITCH_CODE
      }
 #endif
-   
-   
+
    /* loop using overwrite then skip algorithm until done */
    while (1)
      {
@@ -1988,7 +2106,7 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 	       }
 #endif
 
-	     if (CHAR_EQS(q, p)) 
+	     if (CHAR_EQS(q, p))
 	       {
 		  /* Could be the second half of a double width character */
 		  if (p->nchars || q->nchars)
@@ -2002,7 +2120,7 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 	if (buf != buffer) send_attr_str (buffer, buf);
 	buf = buffer;
 
-	if (n_spaces 
+	if (n_spaces
 	    && ((p < pmax) 	       /* erase to eol will achieve this effect*/
 		|| (!CHAR_EQS_SPACE(space_char))))/* unless space_char is not a simple space */
 	  {
@@ -2012,7 +2130,7 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 	 * of a possible match of new and old.  If this is the case, skip
 	 * some more.
 	 */
-	
+
 	/* Note that from here on, the buffer will contain matched characters */
 #if !SLANG_HAS_KANJI_SUPPORT
 	while ((p < pmax) && CHAR_EQS(p, q))
@@ -2077,11 +2195,11 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 	  }
      }
 
-   /* At this point we have reached the end of the new string with the 
+   /* At this point we have reached the end of the new string with the
     * exception of space_chars hanging off the end of it, but we may not have
     * reached the end of the old string if they did not match.
     */
-   
+
    /* Here the buffer will consist only of characters that have matched */
    if (buf != buffer)
      {
@@ -2099,12 +2217,12 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
 	  }
      }
 
-   if (q < qmax) 
+   if (q < qmax)
      {
 	SLtt_reverse_video (COLOR_OF(space_char));
 	del_eol ();
      }
-   
+
 #if SLTT_USE_INSERT_HACK
    else if (insert_hack_char != NULL)
      {
@@ -2117,7 +2235,39 @@ void SLtt_smart_puts(SLsmg_Char_Type *neww, SLsmg_Char_Type *oldd, int len, int
      }
 #endif
 
-   if (Automatic_Margins && (Cursor_c + 1 >= SLtt_Screen_Cols)) Cursor_Set = 0;
+   if (Cursor_c >= SLtt_Screen_Cols)
+     {
+	if (Use_Relative_Cursor_Addressing)
+	  {
+	     /* Where is the cursor?  If we have automatic margins, then the
+	      * cursor _may_ be at the beginning of the next row.  But, it may
+	      * not if the terminal permits writing to the LR corner without
+	      * scrolling.  In relative-cursor-addressing mode, we do not know.
+	      *
+	      * The trick is to force the cursor to the next line if
+	      * it is at the end of the current line, or if it is at
+	      * the beginning of the next do nothing to change that line.
+	      * The only thing I can think of is to write the character that
+	      * is already there.
+	      */
+	     Cursor_c = 0;
+	     Cursor_r++;
+	     if (Cursor_r < SLTT_MAX_SCREEN_ROWS)
+	       {
+		  SLsmg_Char_Type *c = Display_Start_Chars + Cursor_r;
+		  if (c->nchars)
+		    send_attr_str (c, c+1);
+		  else
+		    tt_write (" ", 1);
+	       }
+	     else
+	       tt_write (" ", 1);
+	     tt_write ("\r", 1);
+	     Cursor_c =0;
+	  }
+	else if (Automatic_Margins)
+	  Cursor_Set = 0;
+     }
 }
 
 static void get_color_info (void)
@@ -2174,7 +2324,7 @@ static char *fixup_tgetstr (char *what)
 {
    register char *w, *w1;
    char *wsave;
-   
+
    if (what == NULL)
      return NULL;
 
@@ -2187,7 +2337,7 @@ static char *fixup_tgetstr (char *what)
    while ((*what == '.') ||
 	  ((*what >= '0') && (*what <= '9'))) what++;
    if (*what == '*') what++;
-   
+
    /* lose terminfo padding--- looks like $<...> */
    w = what;
    while (*w) if ((*w++ == '$') && (*w == '<'))
@@ -2215,7 +2365,7 @@ static char *tt_tgetstr (SLCONST char *cap)
 #endif
    if (Termcap_Initalized == 0)
      return NULL;
-   
+
 #ifdef USE_TERMCAP
    /* tmp_area = &Termcap_String_Buf; */
    area = area_buf;
@@ -2243,7 +2393,7 @@ The integrity of this program has been violated.\n");
 	/* It looks like tgetstr placed the object in the buffer and
 	 * returned a pointer to that buffer.  So, we have to make a
 	 * copy of it.
-	 * 
+	 *
 	 * Yes, this introduces a leak...
 	 */
 	s = SLmake_string (s);
@@ -2289,31 +2439,28 @@ int SLtt_tgetflag (SLFUTURE_CONST char *s)
    return tt_tgetflag (s);
 }
 
-#if 0
 int SLtt_tgetent(char *term)
 {
-   return SLtt_initialize(term) == 0;
+   return 0 == SLtt_initialize(term);
 }
 
 int SLtt_tputs(char *str, int affcnt, int (*p)(int))
 {
+   (void) affcnt;
    while (*str) (*p)(*str++);
    return 0;
 }
 
-
-char *SLtt_tgoto(char *cap, int col, int row)
+char *SLtt_tgoto (char *cap, int col, int row)
 {
-  static char buf[64];
-
-  /* beware of overflows. 2^64 is 20 bytes printed */
-  if (strlen(cap) > 23)
-	strcpy(buf, "capability too long");
-  else
-	tt_sprintf(buf, cap, row, col);
-  return buf;
+   static char buf[64];
+   /* beware of overflows. 2^64 is 20 bytes printed */
+   if (strlen(cap) > 23)
+     strcpy(buf, "capability too long");
+   else
+     tt_sprintf(buf, sizeof(buf), cap, row, col);
+    return buf;
 }
-#endif
 
 static int Vt100_Like = 0;
 
@@ -2364,7 +2511,7 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
 	/* Apparantly, this cannot fail according to the man pages. */
 	SLang_TT_Write_FD = fileno (stdout);
      }
-   
+
    if (term == NULL)
      {
 	term = getenv ("TERM");
@@ -2375,7 +2522,7 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
    if (_pSLsecure_issetugid ()
        && ((term[0] == '.') || (NULL != strchr(term, '/'))))
      return -1;
-   
+
    Linux_Console = (!strncmp (term, "linux", 5)
 # ifdef linux
 		    || !strncmp(term, "con", 3)
@@ -2422,7 +2569,7 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
    Termcap_Initalized = 1;
 
    Cls_Str = tt_tgetstr ("cl");
-   Curs_Pos_Str = tt_tgetstr ("cm");
+   Abs_Curs_Pos_Str = tt_tgetstr ("cm");
 
    if ((NULL == (Ins_Mode_Str = tt_tgetstr("im")))
        || ( NULL == (Eins_Mode_Str = tt_tgetstr("ei")))
@@ -2430,7 +2577,28 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
      SLtt_Term_Cannot_Insert = 1;
 
    Visible_Bell_Str = tt_tgetstr ("vb");
+
    Curs_Up_Str = tt_tgetstr ("up");
+   Curs_UpN_Str = tt_tgetstr ("UP");
+
+   /* Avoid \n for moving down because we cannot rely upon the resulting column */
+   Curs_Dn_Str = tt_tgetstr ("do");
+   if ((Curs_Dn_Str != NULL) && (*Curs_Dn_Str == '\n'))
+     Curs_Dn_Str = NULL;
+
+   Curs_DnN_Str = tt_tgetstr ("DO");
+   Curs_Left_Str = tt_tgetstr ("le");
+   if (Curs_Left_Str == NULL)
+     {
+	Curs_Left_Str = tt_tgetstr ("bc");
+	if (Curs_Left_Str == NULL)
+	  Curs_Left_Str = "\b";
+     }
+   Curs_LeftN_Str = tt_tgetstr ("LE");
+   Curs_Right_Str = tt_tgetstr ("nd");
+   Curs_RightN_Str = tt_tgetstr ("RI");
+   Clear_EOS_Str = tt_tgetstr ("cd");
+
    Rev_Scroll_Str = tt_tgetstr("sr");
    Del_N_Lines_Str = tt_tgetstr("DL");
    Add_N_Lines_Str = tt_tgetstr("AL");
@@ -2438,17 +2606,16 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
    /* Actually these are used to initialize terminals that use cursor
     * addressing.  Hard to believe.
     */
-   Term_Init_Str = tt_tgetstr ("ti");
-   Term_Reset_Str = tt_tgetstr ("te");
+   Start_Abs_Cursor_Addressing_Mode = tt_tgetstr ("ti");
+   End_Abs_Cursor_Addressing_Mode = tt_tgetstr ("te");
 
    /* If I do this for vtxxx terminals, arrow keys start sending ESC O A,
     * which I do not want.  This is mainly for HP terminals.
     */
-   if ((almost_vtxxx == 0) || SLtt_Force_Keypad_Init)
-     {
-	Keypad_Init_Str = tt_tgetstr ("ks");
-	Keypad_Reset_Str = tt_tgetstr ("ke");
-     }
+   Keypad_Init_Str = tt_tgetstr ("ks");
+   Keypad_Reset_Str = tt_tgetstr ("ke");
+   if ((almost_vtxxx == 0) && (SLtt_Force_Keypad_Init == -1))
+     SLtt_Force_Keypad_Init = 1;
 
    /* Make up for defective termcap/terminfo databases */
    if ((Vt100_Like && (term[2] != '1'))
@@ -2549,8 +2716,8 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
 	Return_From_Status_Line_Str = tt_tgetstr ("fs");
 	Goto_Status_Line_Str = tt_tgetstr ("ts");
 	/* Status_Line_Esc_Ok = TGETFLAG("es"); */
-	Num_Status_Line_Columns = tt_tgetnum ("ws");
-	if (Num_Status_Line_Columns < 0) Num_Status_Line_Columns = 0;
+	/* Num_Status_Line_Columns = tt_tgetnum ("ws"); */
+	/* if (Num_Status_Line_Columns < 0) Num_Status_Line_Columns = 0; */
      }
 
    if (NULL == (Norm_Vid_Str = tt_tgetstr("me")))
@@ -2561,16 +2728,6 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
    Cursor_Invisible_Str = tt_tgetstr("vi");
    Cursor_Visible_Str = tt_tgetstr("ve");
 
-   Curs_F_Str = tt_tgetstr("RI");
-
-# if 0
-   if (NULL != Curs_F_Str)
-     {
-	Len_Curs_F_Str = strlen(Curs_F_Str);
-     }
-   else Len_Curs_F_Str = strlen(Curs_Pos_Str);
-# endif
-
    Automatic_Margins = TGETFLAG ("am");
    /* No_Move_In_Standout = !TGETFLAG ("ms"); */
 # ifdef HP_GLITCH_CODE
@@ -2588,8 +2745,8 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
      SLtt_Has_Alt_Charset = 0;
 
    Reset_Color_String = tt_tgetstr ("op");
-   
-   /* Apparantly the difference between "AF" and "Sf" is that AF uses RGB, 
+
+   /* Apparantly the difference between "AF" and "Sf" is that AF uses RGB,
     * but Sf uses BGR.
     */
    Color_Fg_Str = tt_tgetstr ("AF"); /* ANSI setaf */
@@ -2637,14 +2794,14 @@ int SLtt_initialize (SLFUTURE_CONST char *term)
 #endif
    get_color_info ();
 
-  
+   TT_Is_Initialized = 1;
+
    if ((Cls_Str == NULL)
-       || (Curs_Pos_Str == NULL))
+       || (Abs_Curs_Pos_Str == NULL))
      return -2;
 
    return 0;
 }
-
 #endif
 /* Unix */
 
@@ -2661,6 +2818,7 @@ void SLtt_enable_cursor_keys (void)
 int SLtt_initialize (char *term)
 {
    SLtt_get_terminfo ();
+   TT_Is_Initialized = 1;
    return 0;
 }
 
@@ -2671,7 +2829,7 @@ void SLtt_get_terminfo ()
    /* Apparantly, this cannot fail according to the man pages. */
    if (SLang_TT_Write_FD == -1)
      SLang_TT_Write_FD = fileno (stdout);
-   
+
    Can_Background_Color_Erase = 0;
 
    Color_Fg_Str = "\033[3%dm";
@@ -2707,9 +2865,17 @@ void SLtt_set_term_vtxxx(int *vt100)
    Del_Eol_Str = "\033[K";
    Del_Bol_Str = "\033[1K";
    Rev_Scroll_Str = "\033M";
-   Curs_F_Str = "\033[%dC";
-   /* Len_Curs_F_Str = 5; */
-   Curs_Pos_Str = "\033[%i%d;%dH";
+
+   Curs_Up_Str = "\033[A";
+   Curs_Dn_Str = "\033[B";
+   Curs_Right_Str = "\033[C";
+   Curs_Left_Str = "\033[D";
+   Curs_UpN_Str = "\033[%dA";
+   Curs_DnN_Str = "\033[%dB";
+   Curs_RightN_Str = "\033[%dC";
+   Curs_LeftN_Str = "\033[%dD";
+
+   Abs_Curs_Pos_Str = "\033[%i%d;%dH";
    if ((vt100 == NULL) || (*vt100 == 0))
      {
 	Ins_Mode_Str = "\033[4h";
@@ -2729,12 +2895,31 @@ void SLtt_set_term_vtxxx(int *vt100)
    /* No_Move_In_Standout = 0; */
 }
 
+void SLtt_init_keypad (void)
+{
+   if (SLtt_Force_Keypad_Init > 0)
+     {
+	tt_write_string (Keypad_Init_Str);
+	SLtt_flush_output ();
+     }
+}
+
+void SLtt_deinit_keypad (void)
+{
+   if (SLtt_Force_Keypad_Init > 0)
+     {
+	tt_write_string (Keypad_Reset_Str);
+	SLtt_flush_output ();
+     }
+}
+
 int SLtt_init_video (void)
 {
    /*   send_string_to_term("\033[?6h"); */
    /* relative origin mode */
-   tt_write_string (Term_Init_Str);
-   tt_write_string (Keypad_Init_Str);
+   if (Use_Relative_Cursor_Addressing == 0)
+     tt_write_string (Start_Abs_Cursor_Addressing_Mode);
+   SLtt_init_keypad ();
    SLtt_reset_scroll_region();
    SLtt_end_insert();
    tt_write_string (Enable_Alt_Char_Set);
@@ -2742,6 +2927,40 @@ int SLtt_init_video (void)
    return 0;
 }
 
+int _pSLtt_init_cmdline_mode (void)
+{
+   if (TT_Is_Initialized == 0)
+     {
+	int status = SLtt_initialize (NULL);
+	if (status < 0)
+	  {
+	     if (status == -1)
+	       SLang_vmessage ("%s", "**WARNING: Unknown terminal capabilities.\n");
+	     return 0;
+	  }
+     }
+   /* We need to be able to use relative cursor addressing in this mode */
+   if (((Curs_UpN_Str == NULL) && (Curs_Up_Str == NULL))
+       || ((Curs_Dn_Str == NULL) && (Curs_DnN_Str == NULL))
+       || ((Curs_Right_Str == NULL) && (Curs_RightN_Str == NULL))
+       || ((Curs_Left_Str == NULL) && (Curs_LeftN_Str == NULL)))
+     return 0;
+
+   SLtt_Term_Cannot_Scroll = 1;
+   SLtt_Use_Ansi_Colors = 0;
+   Use_Relative_Cursor_Addressing = 1;
+   return 1;
+}
+
+void _pSLtt_cmdline_mode_reset (void)
+{
+   Cursor_Set = 0;
+   Cursor_r = Cursor_c = 0;
+   Max_Relative_Cursor_r = 0;
+   tt_write ("\r", 1);
+   _pSLtt_clear_eos ();
+}
+
 int SLtt_reset_video (void)
 {
    SLtt_goto_rc (SLtt_Screen_Rows - 1, 0);
@@ -2764,8 +2983,10 @@ int SLtt_reset_video (void)
 	Current_Fgbg = 0xFFFFFFFFU;
      }
    SLtt_erase_line ();
-   tt_write_string (Keypad_Reset_Str);
-   tt_write_string (Term_Reset_Str);
+   SLtt_deinit_keypad ();
+
+   if (Use_Relative_Cursor_Addressing == 0)
+     tt_write_string (End_Abs_Cursor_Addressing_Mode);
 
    if (Mouse_Mode == 1)
      SLtt_set_mouse_mode (0, 1);
@@ -2825,7 +3046,7 @@ int SLtt_write_to_status_line (SLFUTURE_CONST char *s, int col)
 
 void SLtt_get_screen_size (void)
 {
-#ifdef VMS
+#ifdef VMS_SYSTEM
    int status, code;
    unsigned short chan;
    $DESCRIPTOR(dev_dsc, "SYS$INPUT:");
@@ -2850,7 +3071,7 @@ void SLtt_get_screen_size (void)
 
 #endif
 
-#ifdef VMS
+#ifdef VMS_SYSTEM
    status = sys$assign(&dev_dsc,&chan,0,0,0);
    if (status & 1)
      {
@@ -2899,7 +3120,7 @@ int _pSLtt_get_bce_color_offset (void)
 	else
 	  Bce_Color_Offset = 1;
      }
-   
+
    return Bce_Color_Offset;
 }
 #endif
