@@ -1,6 +1,6 @@
 /* sllimits.h */
 /*
-Copyright (C) 2004, 2005 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -17,7 +17,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #define USE_NEW_HASH_CODE	1
@@ -59,6 +59,9 @@ USA.
 # define SLSTATIC_HASH_TABLE_SIZE	73
 #endif
 
+/* Do not change this */
+#define SL_MAX_TOKEN_LEN		254
+
 /* Size of the keyboard buffer use by the ungetkey routines */
 #ifdef __MSDOS_16BIT__
 # define SL_MAX_INPUT_BUFFER_LEN	40
