@@ -1,7 +1,7 @@
 /* -*- mode: C; mode: fold; -*- */
 /* slnspace.c  --- Name Space implementation */
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -18,7 +18,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #include "slinclud.h"
@@ -46,8 +46,7 @@ SLang_NameSpace_Type *_pSLns_find_namespace (SLCONST char *name)
 }
 
 /* This function deletes the namespace.  It is up to the caller to ensure that
- * the table has already been removed from the list.  Currently, it makes no
- * attempt to delete objects attached to the table.
+ * the table has already been removed from the list.
  */
 void _pSLns_deallocate_namespace (SLang_NameSpace_Type *ns)
 {
@@ -57,8 +56,6 @@ void _pSLns_deallocate_namespace (SLang_NameSpace_Type *ns)
 
    if (ns == NULL)
      return;
-   
-   SLang_free_slstring ((char *) ns->name);
 
    table = ns->table;
    table_size = ns->table_size;
@@ -66,15 +63,17 @@ void _pSLns_deallocate_namespace (SLang_NameSpace_Type *ns)
    for (i = 0; i < table_size; i++)
      {
 	SLang_Name_Type *t = table [i];
-	while (t != NULL)
+   	while (t != NULL)
 	  {
 	     SLang_Name_Type *t1 = t->next;
 	     SLang_free_slstring ((char *) t->name);
 	     SLfree ((char *) t);
-	     /* Later: free additional objects here */
 	     t = t1;
 	  }
      }
+   SLang_free_slstring ((char *) ns->name);
+   SLang_free_slstring ((char *) ns->namespace_name);
+   SLang_free_slstring ((char *) ns->private_name);
    SLfree ((char *)table);
    SLfree ((char *) ns);
 }
@@ -105,7 +104,7 @@ SLang_NameSpace_Type *_pSLns_allocate_namespace (SLFUTURE_CONST char *name, unsi
    ns->name = name;
    ns->table = nt;
    ns->table_size = size;
-   
+
    return ns;
 }
 
@@ -132,8 +131,8 @@ SLang_NameSpace_Type *_pSLns_new_namespace (SLFUTURE_CONST char *name, unsigned
    return table_list;
 }
 
-/* Find the private namespace associated with the object (file, etc) given by 
- * `name', and whose private namespace is given by `namespace_name'.  If 
+/* Find the private namespace associated with the object (file, etc) given by
+ * `name', and whose private namespace is given by `namespace_name'.  If
  * `namespace_name' is NULL, then it is anonymous.
  */
 SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, SLFUTURE_CONST char *namespace_name)
@@ -143,7 +142,7 @@ SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, S
    if ((namespace_name != NULL)
        && (*namespace_name == 0))
      namespace_name = NULL;
-     
+
    ns = Namespace_Tables;
    while (ns != NULL)
      {
@@ -172,7 +171,7 @@ SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, S
 
 	ns = ns->next;
      }
-   
+
    if (NULL == (ns = _pSLns_new_namespace (name, SLSTATIC_HASH_TABLE_SIZE)))
      return ns;
 
@@ -191,14 +190,14 @@ SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, S
 int _pSLns_set_namespace_name (SLang_NameSpace_Type *t, SLFUTURE_CONST char *name)
 {
    SLang_NameSpace_Type *t1;
-   
+
    t1 = _pSLns_find_namespace (name);
    if (t == t1)
      return 0;			       /* already has this name */
 
    if (t1 == NULL)
      t1 = t;
-   
+
    if ((t != t1) || (*name == 0))
      {
 	_pSLang_verror (SL_Namespace_Error, "Namespace \"%s\" already exists",
@@ -218,7 +217,7 @@ int _pSLns_set_namespace_name (SLang_NameSpace_Type *t, SLFUTURE_CONST char *nam
 
    SLang_free_slstring ((char *) t->namespace_name);   /* NULL ok */
    t->namespace_name = name;
-   
+
    return 0;
 }
 
@@ -231,7 +230,7 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
    unsigned int i;
    SLRegexp_Type *reg;
    unsigned int two;
-   
+
    at = NULL;
 
    if ((ns == NULL)
@@ -250,7 +249,7 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
    while (two != 0)
      {
 	two--;
-	
+
 	num_matches = 0;
 	for (i = 0; i < table_size; i++)
 	  {
@@ -265,9 +264,13 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
 		     case SLANG_GVARIABLE:
 		       flags = 8;
 		       break;
-		       
+
 		     case SLANG_ICONSTANT:
 		     case SLANG_DCONSTANT:
+		     case SLANG_FCONSTANT:
+		     case SLANG_LLCONSTANT:
+		     case SLANG_HCONSTANT:
+		     case SLANG_LCONSTANT:
 		     case SLANG_RVARIABLE:
 		     case SLANG_IVARIABLE:
 		       flags = 4;
@@ -280,16 +283,16 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
 		     case SLANG_ARITH_BINARY:
 		       flags = 1;
 		       break;
-		       
+
 		     case SLANG_FUNCTION:
 		       flags = 2;
 		       break;
-		       
+
 		     default:
 		       flags = 0;
 		       break;
 		    }
-		  
+
 		  if ((flags & what)
 		      && (NULL != SLregexp_match (reg, name, strlen (name))))
 		    {
@@ -303,7 +306,7 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
 		  t = t->next;
 	       }
 	  }
-	
+
 	if (at == NULL)
 	  {
 	     at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &num_matches, 1);
@@ -315,7 +318,7 @@ SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *ns, SLFUTURE_CONST c
    if (reg != NULL)
      SLregexp_free (reg);
    return at;
-   
+
    return_error:
    SLregexp_free (reg);
    SLang_free_array (at);
@@ -332,7 +335,7 @@ SLang_NameSpace_Type *_pSLns_create_namespace2 (SLFUTURE_CONST char *name, SLFUT
    ns = _pSLns_find_namespace (namespace_name);
    if (ns != NULL)
      return ns;
-   
+
    if (NULL == (ns = _pSLns_new_namespace (name, SLSTATIC_HASH_TABLE_SIZE)))
      return NULL;
 
@@ -350,10 +353,8 @@ SLang_NameSpace_Type *SLns_create_namespace (SLFUTURE_CONST char *namespace_name
    return _pSLns_create_namespace2 (NULL, namespace_name);
 }
 
-
 void SLns_delete_namespace (SLang_NameSpace_Type *ns)
 {
-   /* FIXME: remove symbols from the namespace */
    if (ns == NULL)
      return;
 
@@ -362,7 +363,7 @@ void SLns_delete_namespace (SLang_NameSpace_Type *ns)
    else
      {
 	SLang_NameSpace_Type *prev = Namespace_Tables;
-	while (prev != NULL) 
+	while (prev != NULL)
 	  {
 	     if (prev->next != ns)
 	       {
@@ -373,7 +374,6 @@ void SLns_delete_namespace (SLang_NameSpace_Type *ns)
 	     break;
 	  }
      }
-
    _pSLns_deallocate_namespace (ns);
 }
 
@@ -382,7 +382,7 @@ SLang_Array_Type *_pSLns_list_namespaces (void)
    SLang_NameSpace_Type *table_list;
    SLang_Array_Type *at;
    SLindex_Type num, i;
-   
+
    num = 0;
    table_list = Namespace_Tables;
    while (table_list != NULL)
@@ -397,7 +397,7 @@ SLang_Array_Type *_pSLns_list_namespaces (void)
 
    table_list = Namespace_Tables;
    i = 0;
-   while ((table_list != NULL) 
+   while ((table_list != NULL)
 	  && (i < num))
      {
 	if (table_list->namespace_name != NULL)
@@ -436,7 +436,6 @@ SLang_Name_Type *
    return t;
 }
 
-
 /* It is up to the caller to make sure that the name is not already in the table */
 int _pSLns_add_hashed_name (SLang_NameSpace_Type *ns, SLang_Name_Type *nt, unsigned long hash)
 {
@@ -458,8 +457,7 @@ SLang_NameSpace_Type *_pSLns_find_object_namespace (SLang_Name_Type *nt)
 
    name = nt->name;
    hash = _pSLcompute_string_hash (name);
-   
-   
+
    ns = Namespace_Tables;
    while (ns != NULL)
      {
@@ -468,7 +466,7 @@ SLang_NameSpace_Type *_pSLns_find_object_namespace (SLang_Name_Type *nt)
 	  {
 	     if (t == nt)
 	       return ns;
-	     
+
 	     t = t->next;
 	  }
 	ns = ns->next;
@@ -482,4 +480,74 @@ SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *ns, SLCONST char *nam
    return _pSLns_locate_hashed_name (ns, name, _pSLcompute_string_hash (name));
 }
 
+static void delete_namespace_objects (SLang_NameSpace_Type *ns)
+{
+   SLang_Name_Type **table = ns->table;
+   unsigned int i, table_size = ns->table_size;
+
+   for (i = 0; i < table_size; i++)
+     {
+	SLang_Name_Type *t = table[i];
+	while (t != NULL)
+	  {
+	     SLang_Name_Type *t1 = t->next;
+	     switch (t->name_type)
+	       {
+		case SLANG_PVARIABLE:
+		case SLANG_GVARIABLE:
+		  SLang_free_object (&((SLang_Global_Var_Type *)t)->obj);
+		  break;
+
+		case SLANG_PFUNCTION:
+		case SLANG_FUNCTION:
+		  SLang_free_function (t);
+		  break;
+
+		case SLANG_ICONSTANT:
+		case SLANG_DCONSTANT:
+		case SLANG_FCONSTANT:
+		case SLANG_LLCONSTANT:
+		case SLANG_HCONSTANT:
+		case SLANG_LCONSTANT:
+		case SLANG_RVARIABLE:
+		case SLANG_IVARIABLE:
+		case SLANG_INTRINSIC:
+		case SLANG_MATH_UNARY:
+		case SLANG_APP_UNARY:
+		case SLANG_ARITH_UNARY:
+		case SLANG_ARITH_BINARY:
+		default:
+		  break;
+	       }
+	     SLang_free_slstring (t->name);
+	     t = t1;
+	  }
+     }
+}
+
+/* This is only called at exit.  Until version 3, I cannot delete everything
+ * since the namespace contains a mixture of statically allocated table
+ * and dynamically allocated ones.  The only way to tell the difference is
+ * to change the API by adding an additional field to SLang_Name_Type object.
+ * So, here I delete what I can safely do, but to avoid leak checkers complaining
+ * about false leaks, the Namespace_Tables pointer will be left as is.
+ */
+void _pSLns_delete_namespaces (void)
+{
+   SLang_NameSpace_Type *ns;
+
+   ns = Namespace_Tables;
+   while (ns != NULL)
+     {
+	SLang_NameSpace_Type *next = ns->next;
+	delete_namespace_objects (ns);
+	SLang_free_slstring ((char *) ns->name);
+	SLang_free_slstring ((char *) ns->namespace_name);
+	SLang_free_slstring ((char *) ns->private_name);
+	/* SLfree ((char *)ns->table); v3*/
+	/* SLfree ((char *) ns); v3 */
+	ns = next;
+     }
+   /* Namespace_Tables = NULL; v3 */
+}
 
