@@ -1,6 +1,6 @@
 /* file intrinsics for S-Lang */
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -17,7 +17,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #include "slinclud.h"
@@ -26,6 +26,15 @@ USA.
 # include <sys/types.h>
 #endif
 
+#include <time.h>
+#ifdef HAVE_SYS_TIME_H
+# include <sys/time.h>
+#endif
+
+#ifdef HAVE_UTIME_H
+# include <utime.h>
+#endif
+
 #ifdef HAVE_IO_H
 # include <io.h>		       /* for chmod */
 #endif
@@ -159,7 +168,7 @@ static SLang_CStruct_Field_Type Stat_Struct [] =
 static int push_stat_struct (struct stat *st, int opt_attrs)
 {
    Stat_Type s;
-   
+
    s.st = *st;
    s.st_opt_attrs = opt_attrs;
    return SLang_push_cstruct ((VOID_STAR) &s, Stat_Struct);
@@ -179,7 +188,7 @@ static void stat_cmd (char *file)
 	file = SLmake_nstring (file, len-1);
 	if (file == NULL)
 	  return;
-	
+
 	is_malloced = 1;
      }
 #endif
@@ -198,12 +207,12 @@ static void stat_cmd (char *file)
 
 	push_stat_struct (&st, opt_attrs);
      }
-   else 
+   else
      {
 	_pSLerrno_errno = errno;
 	SLang_push_null ();
      }
-   
+
 #if defined(__MSDOS__) || defined(__WIN32__)
    if (is_malloced)
      SLfree (file);
@@ -231,13 +240,43 @@ static void lstat_cmd (char *file)
 #else
    opt_attrs = 0;
 #endif
-   
+
    push_stat_struct (&st, opt_attrs);
 #else
    stat_cmd (file);
 #endif
 }
 
+#if defined(HAVE_UTIME) || defined(HAVE_UTIMES)
+static int utime_intrin (char *file, double *t0p, double *t1p)
+{
+# ifdef HAVE_UTIMES
+   struct timeval tv[2];
+# else
+   struct utimbuf u;
+# endif
+   int ret;
+   double t0 = *t0p, t1 = *t1p;
+
+# ifdef HAVE_UTIMES
+   tv[0].tv_sec = (long) t0;
+   tv[0].tv_usec = (long) ((t0 - (long)t0)*1e6);
+   tv[1].tv_sec = (long) t1;
+   tv[1].tv_usec = (long) ((t1 - (long)t1)*1e6);
+   ret = utimes (file, tv);
+# else
+   u.actime = (time_t) t0;
+   u.modtime = (time_t) t1;
+   ret = utime (file, &u);
+# endif
+
+   if (ret == -1)
+     _pSLerrno_errno = errno;
+
+   return ret;
+}
+#endif
+
 /* Well, it appears that on some systems, these are not defined.  Here I
  * provide them.  These are derived from the Linux stat.h file.
  */
@@ -372,7 +411,7 @@ static int symlink_cmd (char *oldpath, char *newpath)
 static int hardlink_cmd (char *oldpath, char *newpath)
 {
    int s;
-   
+
    while (-1 == (s = link (oldpath, newpath)))
      {
 	if (is_interrupt (errno))
@@ -508,7 +547,6 @@ static char *vms_convert_dirspec_to_vms_dir (char *str)
      version = str + len;
    /* version points to the version of the input string */
 
-   
    if (NULL == (s = SLmalloc (len + 8)))/* allow extra space to work with */
      return NULL;
 
@@ -516,7 +554,7 @@ static char *vms_convert_dirspec_to_vms_dir (char *str)
    strncpy (s, str, len);
    s[len] = 0;
    str = s;
-   
+
    /* Lowercase the whole thing */
    while (*s != 0)
      {
@@ -534,11 +572,11 @@ static char *vms_convert_dirspec_to_vms_dir (char *str)
      }
 
    /* Check for one of two possibilities:
-    * 
+    *
     *     dev:[x]   --> dev:x
     *     dev:[a.x] --> dev:[a]x
     */
-   
+
    if (NULL == (dot = strchr (str, '.')))
      {
 	/* First possibility */
@@ -552,13 +590,13 @@ static char *vms_convert_dirspec_to_vms_dir (char *str)
 	*s = 0;
 	goto add_dir_version;
      }
-   
+
    while (NULL != (s = strchr (dot + 1, '.')))
      dot = s;
-   
+
    *dot = ']';
    s = str + (len - 1);
-   
+
    /* Drop */
 
    add_dir_version:
@@ -575,10 +613,10 @@ static int rmdir_cmd (char *s)
 
    if (NULL == (s = vms_convert_dirspec_to_vms_dir (s)))
      return -1;
-   
+
    status = remove_cmd (s);
    SLfree (s);
-   
+
    return status;
 
 #else
@@ -639,7 +677,7 @@ static int mkdir_cmd (void)
    int ret;
    char *dir;
    int mode = 0777;
-   
+
    if (SLang_Num_Function_Args != 1)
      {
 	if (-1 == SLang_pop_integer (&mode))
@@ -671,7 +709,7 @@ static int mkdir_cmd (void)
 	_pSLerrno_errno = errno;
 	break;
      }
-   
+
    SLang_free_slstring (dir);
    return ret;
 }
@@ -737,7 +775,6 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
    rc = DosQueryPathInfo(file, FIL_STANDARD, &status, sizeof(FILESTATUS3));
 # endif
 
-
 # ifdef __WIN32__
    if (status == (DWORD)-1)
      {
@@ -781,7 +818,7 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
 
    num = 0;
    max_num = 50;
-   list = (char **)SLmalloc (max_num * sizeof(char *));
+   list = (char **)_SLcalloc (max_num, sizeof(char *));
    if (list == NULL)
      {
 	SLfree (pat);
@@ -813,7 +850,7 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
 	     return -1;
 	  }
      }
-# endif   
+# endif
    else while (1)
      {
 	/* Do not include hidden files in the list.  Also, do not
@@ -978,7 +1015,7 @@ static void listdir_cmd (char *dir, char *opt)
 	SLang_push_null ();
 	return;
      }
-   /* If max_num_files == 0, then num_files == 0 and list == NULL.  
+   /* If max_num_files == 0, then num_files == 0 and list == NULL.
     * The realloc step below will malloc list for us.
     */
    if (num_files + 1 < max_num_files)
@@ -1052,12 +1089,12 @@ static int umask_cmd (int *u)
 static int access_cmd (char *path, int *modep)
 {
    int mode = *modep & (R_OK|W_OK|X_OK|F_OK);
-   
+
    while (-1 == access (path, mode))
      {
 	if (is_interrupt (errno))
 	  continue;
-	
+
 	_pSLerrno_errno = errno;
 	return -1;
      }
@@ -1100,6 +1137,9 @@ static SLang_Intrin_Fun_Type PosixDir_Name_Table [] =
 #endif
 #if USE_LISTDIR_INTRINSIC
    MAKE_INTRINSIC("listdir", listdir_cmd_wrap, SLANG_VOID_TYPE, 0),
+#endif
+#ifdef HAVE_UTIME
+   MAKE_INTRINSIC_3("utime", utime_intrin, SLANG_INT_TYPE, SLANG_STRING_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE),
 #endif
    SLANG_END_INTRIN_FUN_TABLE
 };
