@@ -1,6 +1,6 @@
 /* time related system calls */
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -17,7 +17,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 #include "slinclud.h"
@@ -57,6 +57,32 @@ USA.
 /* Turns on/off intrinsics that _may_ get added in the future */
 #define USE_FROM_FUTURE		0
 
+#if ((SIZEOF_LONG < 8) && defined(HAVE_LONG_LONG))
+# define SLANG_TIME_T_TYPE SLANG_LLONG_TYPE
+typedef long long _sltime_t;
+# define PUSH_TIME_T SLang_push_long_long
+# define POP_TIME_T SLang_pop_long_long
+#else
+# define SLANG_TIME_T_TYPE SLANG_LONG_TYPE
+typedef long _sltime_t;
+# define PUSH_TIME_T SLang_push_long
+# define POP_TIME_T SLang_pop_long
+#endif
+
+static int pop_time_t (time_t *tt)
+{
+   _sltime_t t;
+   if (-1 == POP_TIME_T(&t))
+     return -1;
+   *tt = (time_t)t;
+   return 0;
+}
+
+static int push_time_t (time_t t)
+{
+   return PUSH_TIME_T ((_sltime_t) t);
+}
+
 #if defined(IBMPC_SYSTEM)
 /* For other system (Unix and VMS), _pSLusleep is in sldisply.c */
 int _pSLusleep (unsigned long s)
@@ -88,11 +114,27 @@ unsigned int sleep (unsigned int seconds)
 }
 #endif
 
-static char *ctime_cmd (long *tt)
+/* Older POSIX standards had a different interfaces for these.  Avoid them. */
+#if !defined(_POSIX_C_SOURCE) || (_POSIX_C_SOURCE < 199506L)
+# undef HAVE_CTIME_R
+# undef HAVE_GMTIME_R
+# undef HAVE_LOCALTIME_R
+#endif
+
+static char *ctime_cmd (void)
 {
    char *t;
-
-   t = ctime ((time_t *) tt);
+   time_t tt;
+#ifdef HAVE_CTIME_R
+   static char buf[64];
+#endif
+   if (-1 == pop_time_t (&tt))
+     return NULL;
+#ifdef HAVE_CTIME_R
+   t = ctime_r (&tt, buf);
+#else
+   t = ctime (&tt);
+#endif
    t[24] = 0;  /* knock off \n */
    return (t);
 }
@@ -107,7 +149,7 @@ static void sleep_cmd (void)
    if (-1 == SLang_pop_double (&x))
      return;
 
-   if (x < 0.0) 
+   if (x < 0.0)
      x = 0.0;
    secs = (unsigned int) x;
    sleep (secs);
@@ -121,9 +163,9 @@ static void sleep_cmd (void)
 #endif
 }
 
-static long _time_cmd (void)
+static void _time_cmd (void)
 {
-   return (long) time (NULL);
+   (void) push_time_t (time(NULL));
 }
 
 #if defined(__GO32__)
@@ -225,6 +267,8 @@ static int validate_tm (struct tm *tms)
 
 static int push_tm_struct (struct tm *tms)
 {
+   if (tms == NULL)
+     return SLang_push_null ();
    return SLang_push_cstruct ((VOID_STAR) tms, TM_Struct);
 }
 
@@ -236,35 +280,249 @@ static int pop_tm_struct (struct tm *tms)
    memset ((char *) tms, 0, sizeof (*tms));
    if (-1 == SLang_pop_cstruct (tms, TM_Struct))
      return -1;
-   
+
    return validate_tm (tms);
 }
 
-static void localtime_cmd (long *t)
-{
-   time_t tt = (time_t) *t;
-   (void) push_tm_struct (localtime (&tt));
-}
-   
-static void gmtime_cmd (long *t)
-{
 #ifdef HAVE_GMTIME
-   time_t tt = (time_t) *t;
-   (void) push_tm_struct (gmtime (&tt));
+static int call_gmtime (time_t t, struct tm *tmp)
+{
+   struct tm *_tmp;
+# ifdef HAVE_GMTIME_R
+   if (NULL != (_tmp = gmtime_r (&t, tmp)))
+     return 0;
+# else
+   _tmp = gmtime (&t);
+   if (_tmp != NULL)
+     {
+	*tmp = *_tmp;
+	return 0;
+     }
+# endif
+   SLang_verror (SL_RunTime_Error, "libc gmtime returned NULL");
+   return -1;
+}
+#endif
+
+static int call_localtime (time_t t, struct tm *tmp)
+{
+   struct tm *_tmp;
+#ifdef HAVE_LOCALTIME_R
+   if (NULL != (_tmp = localtime_r (&t, tmp)))
+     return 0;
 #else
-   localtime_cmd (t);
+   _tmp = localtime (&t);
+   if (_tmp != NULL)
+     {
+	*tmp = *_tmp;
+	return 0;
+     }
+#endif
+   SLang_verror (SL_RunTime_Error, "libc localtime returned NULL");
+   return -1;
+}
+
+
+static void localtime_cmd (void)
+{
+   time_t t;
+   struct tm tm;
+
+   if (-1 == pop_time_t (&t))
+     return;
+   if (0 == call_localtime (t, &tm))
+     (void) push_tm_struct (&tm);
+}
+
+static void gmtime_cmd (void)
+{
+   time_t t;
+   struct tm tm;
+
+   if (-1 == pop_time_t (&t))
+     return;
+
+#ifdef HAVE_GMTIME
+   if (0 == call_gmtime (t, &tm))
+     (void) push_tm_struct (&tm);
+#else
+   if (0 == call_localtime (t, &tm))
+     (void) push_tm_struct (&tm);
 #endif
 }
 
-#ifdef HAVE_MKTIME
-static long mktime_cmd (void)
+#ifdef HAVE_GMTIME
+/* Implement gmtime using a binary search.  The linux manpage says to
+ * play games with the TZ environment variable but that is too ugly for
+ * my taste.
+ */
+static int tm_cmp (struct tm *a, struct tm *b)
 {
-   struct tm t;
+   if (a->tm_year != b->tm_year)
+     return a->tm_year - b->tm_year;
+   if (a->tm_yday != b->tm_yday)
+     return a->tm_yday - b->tm_yday;
+   if (a->tm_hour - b->tm_hour)
+     return a->tm_hour - b->tm_hour;
+   return (a->tm_min - b->tm_min)*60 + (a->tm_sec - b->tm_sec);
+}
 
-   if (-1 == SLang_pop_cstruct (&t, TM_Struct))
-     return (long)-1;
-   
-   return (long) mktime (&t);
+static int timegm_internal (struct tm *tmp, time_t *tp)
+{
+   struct tm tm0;
+   struct tm tm1;
+   time_t t0, t1, t, dt;
+   static time_t delta;
+   static int delta_ok = 0;
+   int diff;
+
+   /* Construct t0 and t1 to bracket the answer.  First, guess a value close
+    * to the correct answer.  Consider:
+    *    tm_utc = gmtime(t)
+    *    tm_loc = localtime(t)
+    *    delta = tm_loc - tm_utc
+    * Then we expect
+    *    t = mktime (tm_loc) = timegm(tm_utc)
+    *      = mktime (tm_utc + delta)
+    *      = mktime (tm_utc) + delta
+    * ==> timegm(tm) = mktime(tm) + delta
+    * ==> tm = gmtime(mktime(tm+delta))
+    *        = gmtime(mktime(tm)) + delta
+    *    delta = tm - gmtime(mktime(tm))
+    */
+   if (delta_ok == 0)
+     {
+	/* localtime Jan 15, 2000 00:00:00 */
+	memset ((char *)&tm1, 0, sizeof(tm0));
+	tm1.tm_mday = 15;
+	tm1.tm_year = 100;
+
+	/* want mktime(tm1), but mktime will modify tm1, so use tm0 as a tmp */
+	tm0 = tm1; t1 = mktime (&tm0);
+	if (-1 == call_gmtime (t1, &tm0))
+	  return -1;
+	/* Compute delta = tm1 - tm0 */
+	tm1.tm_hour += 24*(tm1.tm_mday - tm0.tm_mday);
+	delta = (tm1.tm_hour - tm0.tm_hour)*3600
+	  + (tm1.tm_min - tm0.tm_min)*60 + (tm1.tm_sec - tm0.tm_sec);
+	delta_ok = 1;
+     }
+
+   tm0 = *tmp;			       /* do not allow mktime to mess with *tmp */
+   t = mktime (&tm0) + delta;
+
+   /* Find a lower bound */
+   dt = 0;
+   while (1)
+     {
+	int wrapped = 0;
+	t0 = t - dt;
+	while (t0 > t)/* handle wrapping */
+	  {
+	     wrapped = 1;
+	     t0++;
+	  }
+	if (-1 == call_gmtime (t0, &tm0))
+	  return -1;
+	diff = tm_cmp (&tm0, tmp);
+	if (diff == 0)
+	  {
+	     *tp = t0;
+	     return 0;
+	  }
+	if (diff < 0)
+	  break;
+
+	if (wrapped)
+	  {
+	     SLang_verror (SL_Internal_Error, "timegm: Unable to find a lower limit");
+	     return -1;
+	  }
+
+	dt = dt*2 + 1;
+     }
+
+   /* upper bound */
+   dt = 1;
+   while (1)
+     {
+	int wrapped = 0;
+	t1 = t + dt;
+	while (t1 < t)/* handle wrapping */
+	  {
+	     wrapped = 1;
+	     t1--;
+	  }
+	if (-1 == call_gmtime (t1, &tm1))
+	  return -1;
+	diff = tm_cmp (&tm1, tmp);
+	if (diff == 0)
+	  {
+	     *tp = t1;
+	     return 0;
+	  }
+	if (diff>0)
+	  break;
+
+	if (wrapped)
+	  {
+	     SLang_verror (SL_Internal_Error, "timegm: Unable to find an upper limit");
+	     return -1;
+	  }
+	dt = dt * 2;
+     }
+
+   do
+     {
+	struct tm tmx;
+	t = t0 + (t1-t0)/2;
+	if (-1 == call_gmtime (t, &tmx))
+	  return -1;
+	diff = tm_cmp (&tmx, tmp);
+	if (diff == 0)
+	  {
+	     *tp = t;
+	     return 0;
+	  }
+	if (diff < 0)
+	  {
+	     if (t0 == t)
+	       break;
+	     t0 = t; tm0 = tmx;
+	  }
+	else
+	  {
+	     t1 = t; tm1 = tmx;
+	  }
+     }
+   while (t0 + 1 < t1);
+   *tp = t1;
+   return 0;
+}
+
+static void timegm_cmd (void)
+{
+   struct tm tm;
+   time_t t;
+
+   if (-1 == pop_tm_struct (&tm))
+     return;
+   if (-1 == timegm_internal (&tm, &t))
+     return;
+   (void) push_time_t (t);
+}
+
+#endif
+
+#ifdef HAVE_MKTIME
+static void mktime_cmd (void)
+{
+   struct tm tm;
+
+   if (-1 == SLang_pop_cstruct (&tm, TM_Struct))
+     return;
+
+   (void) push_time_t (mktime (&tm));
 }
 #endif
 
@@ -278,7 +536,7 @@ static long mktime_cmd (void)
 
 #ifdef CLK_TCK
 # define SECS_PER_TICK (1.0/(double)CLK_TCK)
-#else 
+#else
 # ifdef CLOCKS_PER_SEC
 #  define SECS_PER_TICK (1.0/(double)CLOCKS_PER_SEC)
 # else
@@ -287,7 +545,7 @@ static long mktime_cmd (void)
 #endif
 
 typedef struct
-{ 
+{
    double tms_utime;
    double tms_stime;
    double tms_cutime;
@@ -311,7 +569,7 @@ static void times_cmd (void)
 
    (void) times (&t);
 
-   d.tms_utime = SECS_PER_TICK * (double)t.tms_utime; 
+   d.tms_utime = SECS_PER_TICK * (double)t.tms_utime;
    d.tms_stime = SECS_PER_TICK * (double)t.tms_stime;
    d.tms_cutime = SECS_PER_TICK * (double)t.tms_cutime;
    d.tms_cstime = SECS_PER_TICK * (double)t.tms_cstime;
@@ -331,7 +589,7 @@ static double _toc_cmd (void)
    double d;
 
    (void) times (&t);
-   
+
    d = ((t.tms_utime - Tic_TMS.tms_utime)
 	+ (t.tms_stime - Tic_TMS.tms_stime)) * SECS_PER_TICK;
    Tic_TMS = t;
@@ -382,7 +640,7 @@ static double toc_cmd (void)
 # endif
 #endif
 }
-     
+
 #ifdef SLANG_HAS_FLOAT
 # if USE_FROM_FUTURE
 static double ftime_cmd (void)
@@ -424,29 +682,30 @@ static void strftime_cmd (void)
    if (SLang_Num_Function_Args == 1)
      {
 	time_t t = time(NULL);
-	tms = *localtime(&t);
+	if (-1 == call_localtime (t, &tms))
+	  return;
 	if (-1 == validate_tm (&tms))
 	  return;
      }
    else if (-1 == pop_tm_struct (&tms))
      return;
-   
+
    if (-1 == SLang_pop_slstring (&fmt))
      return;
 
    /* Ugh.  The man page says:
-    * 
+    *
     *  The strftime() function returns the number of characters placed in  the
     *  array  s,  not  including  the  terminating NUL character, provided the
     *  string, including the terminating NUL, fits.  Otherwise, it returns  0,
     *  and  the contents of the array is undefined.  (Thus at least since libc
     *  4.4.4; very old versions of libc, such as libc 4.4.1, would return  max
     *  if the array was too small.)
-    *  
+    *
     *  Note  that  the  return value 0 does not necessarily indicate an error;
     *  for example, in many locales %p yields an empty string.
-    * 
-    * Was this too designed by committee? 
+    *
+    * Was this too designed by committee?
     */
    status = strftime (buf, sizeof(buf), fmt, &tms);
    if (status == 0)
@@ -458,14 +717,17 @@ static void strftime_cmd (void)
 
 static SLang_Intrin_Fun_Type Time_Funs_Table [] =
 {
-   MAKE_INTRINSIC_1("ctime", ctime_cmd, SLANG_STRING_TYPE, SLANG_LONG_TYPE),
+   MAKE_INTRINSIC_0("ctime", ctime_cmd, SLANG_STRING_TYPE),
    MAKE_INTRINSIC_0("sleep", sleep_cmd, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("_time", _time_cmd, SLANG_LONG_TYPE),
+   MAKE_INTRINSIC_0("_time", _time_cmd, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("time", SLcurrent_time_string, SLANG_STRING_TYPE),
-   MAKE_INTRINSIC_1("localtime", localtime_cmd, SLANG_VOID_TYPE, SLANG_LONG_TYPE),
-   MAKE_INTRINSIC_1("gmtime", gmtime_cmd, SLANG_VOID_TYPE, SLANG_LONG_TYPE),
+   MAKE_INTRINSIC_0("localtime", localtime_cmd, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("gmtime", gmtime_cmd, SLANG_VOID_TYPE),
+#ifdef HAVE_GMTIME
+   MAKE_INTRINSIC_0("timegm", timegm_cmd, SLANG_VOID_TYPE),
+#endif
 #ifdef HAVE_MKTIME
-   MAKE_INTRINSIC_0("mktime", mktime_cmd, SLANG_LONG_TYPE),
+   MAKE_INTRINSIC_0("mktime", mktime_cmd, SLANG_VOID_TYPE),
 #endif
    MAKE_INTRINSIC_0("strftime", strftime_cmd, SLANG_VOID_TYPE),
 #ifdef HAVE_TIMES
@@ -485,6 +747,10 @@ static SLang_Intrin_Fun_Type Time_Funs_Table [] =
 
 int _pSLang_init_sltime (void)
 {
+   if (sizeof (time_t) > sizeof(_sltime_t))
+     {
+	SLang_exit_error ("S-Lang library not built properly.  Fix _sltime_t in sltime.c");
+     }
 #ifdef HAVE_TIMES
    (void) tic_cmd ();
 #endif
