@@ -1,6 +1,6 @@
 /* -*- mode: C; mode: fold -*- */
 /*
-Copyright (C) 2004-2009 John E. Davis
+Copyright (C) 2004-2011 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -17,7 +17,7 @@ General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
+USA.
 */
 
 /* This file is best edited with a folding editor */
@@ -31,10 +31,10 @@ USA.
 #include "slang.h"
 #include "_slang.h"
 
-int SLtt_Term_Cannot_Insert;
-int SLtt_Term_Cannot_Scroll;
+int SLtt_Term_Cannot_Insert = 0;
+int SLtt_Term_Cannot_Scroll = 0;
 int SLtt_Ignore_Beep = 3;
-int SLtt_Use_Ansi_Colors;
+int SLtt_Use_Ansi_Colors = 0;
 int SLtt_Has_Status_Line = 0;
 int SLtt_Screen_Rows = 25;
 int SLtt_Screen_Cols = 80;
@@ -2217,7 +2217,7 @@ int SLtt_set_cursor_visibility (int show)
 {
 #if defined(WIN32_VIDEO)
    CONSOLE_CURSOR_INFO c;
-     
+
    if (0 == GetConsoleCursorInfo (hStdout, &c))
      return -1;
    c.bVisible = (show ? TRUE: FALSE);
@@ -2341,10 +2341,10 @@ int SLtt_set_color (int obj, SLFUTURE_CONST char *what, SLFUTURE_CONST char *fg,
      been set yet, set it to the new color */
    if ((obj == 0) && (Attribute_Byte == 0))
      SLtt_reverse_video (0);
-   
+
    if (_pSLtt_color_changed_hook != NULL)
      (*_pSLtt_color_changed_hook)();
-   
+
    return 0;
 }
 
@@ -2361,7 +2361,6 @@ static void fixup_colors (void)
    SLtt_normal_video ();
 }
 
-
 /* FIXME!!! Add mono support.
  * The following functions have not been fully implemented.
  */
@@ -2385,7 +2384,7 @@ int SLtt_utf8_enable (int mode)
 {
    /* FIXME: UTF-8 mode is disabled until the drivers support UTF-8 */
    mode = 0;
-   
+
    if (mode == -1)
      mode = _pSLtt_UTF8_Mode;
 
